# coding="utf-8"

from sgt import core
import numpy as np
from matplotlib import pyplot

if __name__ == "__main__":
    ax = pyplot.figure().add_subplot(projection="3d")
    ax.set_box_aspect((1,1,1))
    ax.set_xlim(-4000.0, 4000.0)
    ax.set_ylim(-4000.0, 4000.0)
    ax.set_zlim(-4000.0, 4000.0)
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")



    def plot_vector(ax, vec, **kwargs):
        return ax.plot([0.0, vec[0]], [0.0, vec[1]], [0.0, vec[2]], **kwargs)

    sdd = 3193.0

    a, b, n = core.make_basis_vectors_on_detector_in_lab_system(core.make_rotation_matrix(0.0, 0.0, 0.0))
    
    u, v = core.make_pixel_coords_in_detector_system(1475, 1679, 0.172, 0.172, 128.2, 132.7)

    x, y, z = core.make_pixel_coords_in_lab_system(u, v, a, b, n, sdd)

    pyplot.plot([x[0,0], x[0,-1], x[-1,-1], x[-1,0], x[0,0]], 
                [y[0,0], y[0,-1], y[-1,-1], y[-1,0], y[0,0]], 
                [z[0,0], z[0,-1], z[-1,-1], z[-1,0], z[0,0]], 
                marker="", ls="-", color="g")

    pyplot.plot([x[0,0]], 
                [y[0,0]], 
                [z[0,0]], 
                marker="o", ls="", color="g")

    pyplot.plot([0.0, 0.0], 
                [0.0, 0.0], 
                [0.0, sdd], 
                marker="", ls="-", color="k")

    pyplot.show()


# coding="utf-8"

import snpl
import snpl.image
import sgt
from sgt import pilatus, refine
import numpy as np

if __name__ == "__main__":
    Q_AgBh_1 = 2.0*np.pi/58.38

    mask = snpl.image.NpzImage("test data for refine/Mask_EmptyBeam.npz").get_layer("mask")
    geom = sgt.geometry()
    geom.load_specs("test data for refine/Geometry_AgBh.txt")
    geom.mask = mask
    geom.refresh_q()

    geom.specs["u0_mm"] = 73.0
    geom.specs["v0_mm"] = 95.0

    geom.refresh_q()

    q = geom.get_qabs()

    intensity = pilatus.ReadTiff("test data for refine/AgBh.tif").astype(np.float64)

    qtargets = [Q_AgBh_1*1.0, Q_AgBh_1*2.0, Q_AgBh_1*3.0]
    qfuncwidths = [0.005, 0.005, 0.005]
    qlimwidths = [v*4.0 for v in qfuncwidths]

    qfuncmins = [qt - qw for qt, qw in zip(qtargets, qfuncwidths)]
    qfuncmaxs = [qt + qw for qt, qw in zip(qtargets, qfuncwidths)]
    qlimmins = [qt - qw for qt, qw in zip(qtargets, qlimwidths)]
    qlimmaxs = [qt + qw for qt, qw in zip(qtargets, qlimwidths)]

    snpl.figsize(6,6)
    snpl.gca().set_aspect("equal")
    snpl.pyplot.pcolormesh(geom.u, geom.v, intensity, shading="nearest", vmin=0.0, vmax=10000)
    pmask = np.ones_like(mask, dtype=float)
    pmask[mask==0] = np.nan
    snpl.pyplot.pcolormesh(geom.u, geom.v, pmask, shading="nearest", vmin=0, vmax=1, cmap="Greys")

    snpl.pyplot.contour(geom.u, geom.v, q, qtargets, linewidths=0.2, colors="r", linestyles="-")
    snpl.pyplot.contour(geom.u, geom.v, q, qfuncmins, linewidths=0.2, colors="r", linestyles=":")
    snpl.pyplot.contour(geom.u, geom.v, q, qfuncmaxs, linewidths=0.2, colors="r", linestyles=":")
    snpl.pyplot.contour(geom.u, geom.v, q, qlimmins, linewidths=0.1, colors="r", linestyles="-")
    snpl.pyplot.contour(geom.u, geom.v, q, qlimmaxs, linewidths=0.1, colors="r", linestyles="-")


    snpl.axvline(0.0, color="w", lw=0.2)
    snpl.axhline(0.0, color="w", lw=0.2)

    snpl.savefig("test_opt_image.png", dpi=300)
    snpl.clf()


    def callback(ncalls, score):
        print("Calls = {0} Score = {1}".format(ncalls, score))

    refine.refine(geom, 
    refine.score_by_rings, 
    (intensity, qtargets, qlimwidths, qfuncwidths), 
    ("u0_mm", "v0_mm"), 
    callback=callback)

    geom.refresh_q()

    q = geom.get_qabs()

    snpl.gca().set_aspect("equal")
    snpl.pyplot.pcolormesh(geom.u, geom.v, intensity, shading="nearest", vmin=0.0, vmax=10000)
    pmask = np.ones_like(mask, dtype=float)
    pmask[mask==0] = np.nan
    snpl.pyplot.pcolormesh(geom.u, geom.v, pmask, shading="nearest", vmin=0, vmax=1, cmap="Greys")

    snpl.pyplot.contour(geom.u, geom.v, q, qtargets, linewidths=0.2, colors="r", linestyles="-")
    snpl.pyplot.contour(geom.u, geom.v, q, qfuncmins, linewidths=0.2, colors="r", linestyles=":")
    snpl.pyplot.contour(geom.u, geom.v, q, qfuncmaxs, linewidths=0.2, colors="r", linestyles=":")
    snpl.pyplot.contour(geom.u, geom.v, q, qlimmins, linewidths=0.1, colors="r", linestyles="-")
    snpl.pyplot.contour(geom.u, geom.v, q, qlimmaxs, linewidths=0.1, colors="r", linestyles="-")

    snpl.axvline(0.0, color="w", lw=0.2)
    snpl.axhline(0.0, color="w", lw=0.2)

    snpl.savefig("test_opt_image_after.png", dpi=300)


# coding="utf-8"

from sgt import core
import numpy as np
from matplotlib import pyplot

if __name__ == "__main__":
    ax = pyplot.figure().add_subplot(projection="3d")
    ax.set_box_aspect((1,1,1))
    ax.set_xlim(-1.5, 1.5)
    ax.set_ylim(-1.5, 1.5)
    ax.set_zlim(-1.5, 1.5)
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    def plot_vector(ax, vec, **kwargs):
        return ax.plot([0.0, vec[0]], [0.0, vec[1]], [0.0, vec[2]], **kwargs)

    a, b, n = core.make_basis_vectors_on_detector_in_lab_system(core.make_rotation_matrix(0.0, 0.0, 0.0))
    
    plot_vector(ax, a, ls="-", color="r", marker="x", alpha=0.5)
    plot_vector(ax, b, ls="-", color="g", marker="x", alpha=0.5)
    plot_vector(ax, n, ls="-", color="b", marker="x", alpha=0.5)

    a, b, n = core.make_basis_vectors_on_detector_in_lab_system(core.make_rotation_matrix(10.0, 0.0, 0.0))
    
    plot_vector(ax, a, ls="-", color="r", marker="^", alpha=0.5)
    plot_vector(ax, b, ls="-", color="g", marker="^", alpha=0.5)
    plot_vector(ax, n, ls="-", color="b", marker="^", alpha=0.5)

    a, b, n = core.make_basis_vectors_on_detector_in_lab_system(core.make_rotation_matrix(10.0, 10.0, 0.0))
    
    plot_vector(ax, a, ls="-", color="r", marker="s", alpha=0.5)
    plot_vector(ax, b, ls="-", color="g", marker="s", alpha=0.5)
    plot_vector(ax, n, ls="-", color="b", marker="s", alpha=0.5)

    a, b, n = core.make_basis_vectors_on_detector_in_lab_system(core.make_rotation_matrix(10.0, 10.0, 10.0))
    
    plot_vector(ax, a, ls="-", color="r", marker="o", alpha=0.5)
    plot_vector(ax, b, ls="-", color="g", marker="o", alpha=0.5)
    plot_vector(ax, n, ls="-", color="b", marker="o", alpha=0.5)

    pyplot.show()


# coding="utf-8"
"""Higher-level interface for manipulation of scattering geometry. 

"""

from typing import List, IO, Tuple
import numpy as np
import json
from sgt import core, _cpolarize

__version__ = "0.1.1"

class geometry(object):
    """Interface class for scattering geometry manipulation. 

    Example:
        To create a `geometry` instance, 
        
            >>> import sgt
            >>> g = sgt.geometry()
        
        Although all necessary parameters may be supplied 
        as the args of the initializer, 
        the easiest way is to load the specs from a file. 

            >>> g.load_specs("Geometry_AgBh.txt")

        Make sure to call refresh functions 
        when you make changes to the geometry. 
        
            >>> g.refresh_q()
        
        Mask can be supplied later. 
        For example, suppose that we already have the mask array
        `maskarray`, 
        
            >>> g.mask = maskarray

        Make sure to refresh polar maps whenever you made
        any changes to the geometry. 
            
            >>> g.refresh_polar_map()

        Finally, it can perform circular averaging. 
        Suppose that we have the 2D intensity array `i`
        and associated error array `e_i`, 

            >>> i_av, e_i_av = g.circular_average(i, e_i)
        
        The output arrays are also 2D. 
        For example, ``i_av[k]`` is the q-profile at ``k`` th 
        azimuthal section. 
        The arrays along the q and azimuthal angle axes 
        are stored as ``g.ax_q`` and ``g.ax_azi``, 
        respectively. 
    """

    minimal_spec_keys: Tuple = \
        ("width_px", 
         "height_px", 
         "px_width_mm", 
         "px_height_mm", 
         "u0_mm", 
         "v0_mm",   
         "alpha_deg", 
         "beta_deg", 
         "gamma_deg", 
         "L0_mm", 
         "lam_ang", 
         "qmin_anginv", 
         "qmax_anginv", 
         "q_number", 
         "azi_number")

    def __init__(self, 
    width_px: int=1, height_px: int=1, 
    px_width_mm: float=1.0, px_height_mm: float=1.0, 
    u0_mm: float=0.0, v0_mm: float=0.0, 
    alpha_deg: float=0.0, beta_deg: float=0.0, gamma_deg: float=0.0, 
    L0_mm: float=1.0, lam_ang: float=1.0, 
    qmin: float=0.0, qmax: float=1.0, N_q: int=100, N_azi: int=1, 
    mask: np.ndarray|None=None) -> None:

        self.specs: dict = {}

        self.specs["width_px"]     = width_px
        self.specs["height_px"]    = height_px
        self.specs["px_width_mm"]  = px_width_mm
        self.specs["px_height_mm"] = px_height_mm
        self.specs["u0_mm"]        = u0_mm
        self.specs["v0_mm"]        = v0_mm 
        self.specs["alpha_deg"]    = alpha_deg 
        self.specs["beta_deg"]     = beta_deg 
        self.specs["gamma_deg"]    = gamma_deg 
        self.specs["L0_mm"]        = L0_mm 
        self.specs["lam_ang"]      = lam_ang
        self.specs["qmin_anginv"]  = qmin
        self.specs["qmax_anginv"]  = qmax
        self.specs["q_number"]     = N_q
        self.specs["azi_number"]   = N_azi 

        if mask is None:
            self.mask = core.make_default_mask(width_px, height_px)
        else:
            self.mask = mask

        # vars to be calculated
        self._R: np.ndarray = np.zeros((3,3))
        self._u: np.ndarray = np.empty((0,0))
        self._v: np.ndarray = np.empty((0,0))
        self._a: np.ndarray = np.zeros((3,))
        self._b: np.ndarray = np.zeros((3,))
        self._n: np.ndarray = np.zeros((3,))
        self._x: np.ndarray = np.empty((0,0))
        self._y: np.ndarray = np.empty((0,0))
        self._z: np.ndarray = np.empty((0,0))
        self._qx: np.ndarray = np.empty((0,0))
        self._qy: np.ndarray = np.empty((0,0))
        self._qz: np.ndarray = np.empty((0,0))
        self._solid_angle_factor: np.ndarray = np.empty((0,0))
        self._map_q:   np.ndarray = np.empty((0,0), dtype=int)
        self._map_azi: np.ndarray = np.empty((0,0), dtype=int)
        self._density: np.ndarray = np.empty((0,0), dtype=int)
        self._ax_q:    np.ndarray = np.empty((0,0))
        self._ax_azi:  np.ndarray = np.empty((0,0))
        self._normal_incidence_dist: float = 0.0

    def load_specs(self, fp: str|IO) -> None:
        """Loads and applies parameters from a file
        
        It loads parameters from a geometry specification file, 
        which is a JSON-formatted text file. 
        The file must contain all keys listed in ``minimal_spec_keys``. 

        Args:
            fp: file-like or path to a JSON-formatted file. 
        """

        h: dict = {}
        missing_keys: List[str] = []

        if isinstance(fp, str):
            with open(fp, "r") as f:
                lines = f.readlines()
        else:
            lines = fp.readlines()
        
        # for backward compatibility
        # older geometry file contains # at every line head
        if lines[0].startswith("#"):
            jsonfeed = "".join([l.lstrip("#") for l in lines])
        else:
            jsonfeed = "".join(lines)

        h = json.loads(jsonfeed)

        missing_keys = [k for k in self.minimal_spec_keys if k not in h]

        if missing_keys:
            raise ValueError("missing key(s): " + ", ".join(missing_keys))

        self.specs.update(h)
    
    def save_specs(self, fp: str|IO) -> None:
        h: dict = {k: v for k, v in self.specs.items()}
        h["version"] = __version__

        hstr: str = json.dumps(h, indent=4, ensure_ascii=False)

        if isinstance(fp, str):
            with open(fp, "w") as f:
                f.write(hstr)
        else:
            fp.write(hstr)

    def refresh_q(self) -> None:
        """
        
        Note:
            ``self.mask`` is not used in this method. 
        """
        self._R = core.make_rotation_matrix(
            self.specs["alpha_deg"], 
            self.specs["beta_deg"], 
            self.specs["gamma_deg"])

        self._u, self._v = core.make_pixel_coords_in_detector_system(
            self.specs["width_px"], self.specs["height_px"], 
            self.specs["px_width_mm"], self.specs["px_height_mm"], 
            self.specs["u0_mm"], self.specs["v0_mm"]
            )

        self._a, self._b, self._n = core.make_basis_vectors_on_detector_in_lab_system(self._R)

        self._x, self._y, self._z = core.make_pixel_coords_in_lab_system(
            self._u, self._v, 
            self._a, self._b, self._n, self.specs["L0_mm"]
            )

        self._normal_incidence_dist = core.calc_shortest_dist_to_detector(
            self._a, self._b, self._n, self.specs["L0_mm"]
            )

        self._solid_angle_factor = \
            core.make_solid_angle_coverage_correction_factors(
                self._x, self._y, self._z, self._normal_incidence_dist
            )

        self._qx, self._qy, self._qz = core.make_q(
            self._x, self._y, self._z, self.specs["lam_ang"]
            )

    def refresh_polar_map(self) -> None:
        """
        
        Note: 
            This method uses ``self.mask``. 
        """

        assert self._is_ready_for_polar_map() # check

        self._map_q, self._map_azi, self._density, self._ax_q, self._ax_azi = \
            _cpolarize.calc_polar_map(
                self._qx, self._qy, self._qz, 
                self.mask, 
                self.specs["qmin_anginv"], self.specs["qmax_anginv"], 
                self.specs["q_number"], self.specs["azi_number"]
            )
    
    def circular_average(self, intensity: np.ndarray, e_intensity: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Performs circular averaging. 

        Args:
            intensity: intensity array. 
            e_intensity: intensity error array. 
        
        Returns:
            Two 2D numpy arrays of the intensity and the error, 
            both with the shape (``azi_number``, ``q_number``). 
        """

        return _cpolarize.circular_average(
            intensity.astype(np.float64), 
            e_intensity.astype(np.float64), 
            self._map_q, self._map_azi, self._density
            )
    
    def _is_ready_for_polar_map(self) -> bool:

        shape: Tuple[int, int] = (self.specs["height_px"], self.specs["width_px"])
        
        # check mask size
        if self.mask.shape != shape:
            return False
        
        return True

    def get_qabs(self) -> np.ndarray: 
        return np.sqrt(self.qx*self.qx + self.qy*self.qy + self.qz*self.qz)

    @property
    def R(self) -> np.ndarray: return self._R

    @property
    def u(self) -> np.ndarray: return self._u

    @property
    def v(self) -> np.ndarray: return self._v

    @property
    def x(self) -> np.ndarray: return self._x

    @property
    def y(self) -> np.ndarray: return self._y

    @property
    def z(self) -> np.ndarray: return self._z

    @property
    def qx(self) -> np.ndarray: return self._qx

    @property
    def qy(self) -> np.ndarray: return self._qy

    @property
    def qz(self) -> np.ndarray: return self._qz

    @property
    def solid_angle_factor(self) -> np.ndarray: return self._solid_angle_factor

    @property
    def map_q(self) -> np.ndarray: return self._map_q

    @property
    def map_azi(self) -> np.ndarray: return self._map_azi

    @property
    def density(self) -> np.ndarray: return self._density

    @property
    def ax_q(self) -> np.ndarray: return self._ax_q

    @property
    def ax_azi(self) -> np.ndarray: return self._ax_azi

    @property
    def normal_incidence_dist(self) -> float: return self._normal_incidence_dist

if __name__ == "__main__":
    pass


# coding="utf-8"
"""Core tools for scattering geometry calculations. 
"""

from typing import List, IO, Tuple
import numpy as np

import math

from sgt import _cpolarize

def make_rotation_matrix(alpha_deg: float, beta_deg: float, gamma_deg: float) -> np.ndarray:
    """Makes a rotation matrix that defines orientation of the detector. 

    The rotation matrix is calculated based on the given Euler angles. 
    The Euler angles here is the x-y-x type, that is, 
    
    1. The initial axes :math:`(x, y, z)` are rotated by `alpha_deg` around the :math:`x` axis. 
    2. The resultant axes :math:`(x', y', z')` are rotated by `beta_deg` around the :math:`y'` axis. 
    3. The resultant axes :math:`(x'', y'', z'')` are rotated by `gamma_deg` around the :math:`x''` axis. 

    Args:
        alpha_deg: Rotation around the :math:`x` axis, in degrees. 
        beta_deg: Rotation around the :math:`y'` axis, in degrees. 
        gamma_deg: Rotation around the :math:`x''` axis, in degrees. 

    Returns:
        A 3x3 numpy array. 

    Example: 
        >>> R = make_rotation_matrix(10.0, 10.0, 10.0)
        >>> vec = np.array([1.0, 2.0, 3.0])
        >>> rotated = np.matmul(R, vec)
    """
    r1: float = math.radians(alpha_deg)
    r2: float = math.radians(beta_deg)
    r3: float = math.radians(gamma_deg)
    
    c1: float = math.cos(r1)
    c2: float = math.cos(r2)
    c3: float = math.cos(r3)
    s1: float = math.sin(r1)
    s2: float = math.sin(r2)
    s3: float = math.sin(r3)
    
    return np.array([[c2,     s2*s3,          c3*s2], 
                     [s1*s2,  c1*c3-c2*s1*s3, -c1*s3-c2*c3*s1], 
                     [-c1*s2, c3*s1+c1*c2*s3, c1*c2*c3-s1*s3]]) # xyx

def make_pixel_coords_in_detector_system(
    hor_px_num: int, ver_px_num: int, 
    px_width: float, px_height: float, 
    center_coord_hor: float, center_coord_ver: float
    ) -> Tuple[np.ndarray, np.ndarray]:
    """Makes the matrices of coordinates of each pixel on the 2D detector coordinate system. 

    The detector coordinate system is a 2D Cartesian coordinate system 
    whose origin is at the image center (= where the direct beam hits the detector plane). 
    Two axes, denoted as u and v, are defined to be parallel to the horizontal and vertical 
    edges of the detector, respectively. 

    Args:
        hor_px_num: Number of pixels along the horizontal axis. 
        ver_px_num: Number of pixels along the vertical axis. 
        px_width: Size of a single pixel along the horizontal axis. 
        px_height: Size of a single pixel along the horizontal axis. 
        center_coord_hor: Horizontal coordinate of the image center
            measured from the center of pixel at index `[0,0]`. 
        center_coord_ver: Vertical coordinate of the image center. 

    Returns:
        2D numpy arrays of the horizontal and vertical coordinates. 

    Example:
        >>> u, v = make_pixel_coords_in_detector_system(1475, 1679, 0.172, 0.172, 132.35, 134.39)
    """
    # u coordinates = (index of the pixel along u axis)*(pixel width) - (u coord at the center)
    u: np.ndarray = np.arange(hor_px_num).astype(float)*px_width - center_coord_hor

    # v coordinates = (index of the pixel along v axis)*(pixel width) - (v coord at the center)
    v: np.ndarray = np.arange(ver_px_num).astype(float)*px_height - center_coord_ver

    # matrix of u coordinates & v coordinates
    uu: np.ndarray = np.array([])
    vv: np.ndarray = np.array([])
    uu, vv = np.meshgrid(u, v, indexing="xy")

    return uu, vv

def make_default_mask(hor_px_num: int, ver_px_num: int) -> np.ndarray:
    """Makes a default mask array. 

    A mask array is a 2D array of the same shape as the scattering image
    but of `numpy.uint8` type. Pixels to be masked are assigned with 1 
    and unmasked pixels are assigned with zero. 

    Args:
        hor_px_num: Number of pixels along the horizontal edge. 
        ver_px_num: Number of pixels along the horizontal edge. 

    Returns:
        A 2D numpy array of the dtype `numpy.uint8`. 
    """
    return np.zeros((ver_px_num, hor_px_num), dtype=np.uint8)

def make_default_array(hor_px_num: int, ver_px_num: int) -> np.ndarray:
    """Makes a default float array. 

    Args:
        hor_px_num: Number of pixels along the horizontal edge. 
        ver_px_num: Number of pixels along the horizontal edge. 

    Returns:
        A 2D numpy array of the float type. 
    """
    return np.zeros((ver_px_num, hor_px_num), dtype=float)

def make_basis_vectors_on_detector_in_lab_system(rotation_matrix: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Makes the basis vectors of the detector coordinate system, 
    expressed in the lab coordinate system. 

    For the definition of the detector coordinate system, 
    refer to :py:func:`make_pixel_coords_in_detector_system`. 
    The basis vectors of the detector coordinate system 
    is basically the basis vectors of the lab coordinate system
    being rotated by the rotation matrix that defines 
    the detector orientation. 
    The input rotation matrix can be created by 
    :py:func:`make_rotation_matrix`. 
    (but actually can be any SO(3) matrix)

    Args:
        rotation_matrix: A 3x3 numpy array representing 
            the detector orientation. 

    Returns:
        Three numpy arrays `a`, `b`, and `n` representing the basis vectors. 
        `a` and `b` are the basis vector along the horizontal and vertical
        edge of the detector, respectively, and `n` is the one 
        perpendicular to the detector plane. 
    """

    # basis vectors on the detector plane, expressed in the lab coordinate system
    a: np.ndarray = np.matmul(rotation_matrix, np.array([1.0, 0.0, 0.0])) # in-plane basis vector of the detector plane
    b: np.ndarray = np.matmul(rotation_matrix, np.array([0.0, 1.0, 0.0])) # in-plane basis vector of the detector plane
    n: np.ndarray = np.matmul(rotation_matrix, np.array([0.0, 0.0, 1.0])) # plane normal

    return a, b, n


def make_pixel_coords_in_lab_system(
    xcoords_det: np.ndarray, ycoords_det: np.ndarray, 
    a: np.ndarray, b: np.ndarray, n: np.ndarray, sdd: float
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    x: np.ndarray = xcoords_det*a[0] + ycoords_det*b[0]
    y: np.ndarray = xcoords_det*a[1] + ycoords_det*b[1]
    z: np.ndarray = xcoords_det*a[2] + ycoords_det*b[2] + sdd

    return x, y, z

def calc_shortest_dist_to_detector(a: np.ndarray, b: np.ndarray, n: np.ndarray, sdd: float) -> float:
    """Computes the shortest distance from the sample to the detector plane. 

    Let P be the point on the detector such that OP is the shortest distance 
    between the origin and the detector plane. 
    The vector OP must be perpendicular to the detector plane, 
    so the vector :math:`\\vec{\\mathrm{OP}}` is proportional 
    to the detector plane normal vector :math:`\\vec{n}`. 
    That is, 
    
    .. math::
       \\vec{\\mathrm{OP}} = \\mathrm{OP} \\vec{n}

    The vector OP can also be expressed as

    .. math::
       \\vec{\\mathrm{OP}} = u\\vec{a} + v\\vec{b} + L\\vec{e}_z

    where :math:`(u, v)` is the coordinate of point P 
    on the detector coordinate system
    and vector :math:`\\vec{e}_z` is the z basis vector. 
    :math:`L` is the sample-to-detector distance. 

    Equating the two expressions, 

    .. math::
       k\\vec{n} = u\\vec{a} + v\\vec{b} + L\\vec{e}_z

    which reads

    .. math::
       u a_x + v b_x - \\mathrm{OP} n_x &= 0 \\\\
       u a_y + v b_y - \\mathrm{OP} n_y &= 0 \\\\
       u a_z + v b_z - \\mathrm{OP} n_z &= -L
    
    By defining a matrix

    .. math::
       \\mathbf{M} = 
       \\begin{pmatrix}
       a_x & b_x & -n_x \\\\
       a_y & b_y & -n_y \\\\
       a_z & b_z & -n_z \\\\
       \\end{pmatrix}

    the equations are simplified to

    .. math::
       \\mathbf{M} \\vec{s} &= -L \\vec{e}_z \\\\
       \\vec{s} &= -L \\mathbf{M}^{-1}\\vec{e}_z

    where :math:`\\vec{s} = (u, v, \\mathrm{OP})`. 
    This method computes :math:`\\vec{s}` using the above equation
    and returns its third component, :math:`\mathrm{OP}`. 

    Args:
        a: basis vector of the detector coordinate system 
            along the horizontal edge of the detector. 
        b: basis vector of the detector coordinate system
            along the vertical edge of the detector.  
        n: basis vector of the detector coordinate system
            along the plane normal of the detector. 
        sdd: sample-to-detector distance. 
    
    Returns:
        The distance in the float value. 
    
    Note: 
        The input vectors should be expressed 
        in the lab coordinate system. 
        Use :py:func:`make_basis_vectors_on_detector_in_lab_system`
        to generate the basis vectors. 
    """

    M: np.ndarray = np.array([[a[0], b[0], -n[0]],
                              [a[1], b[1], -n[1]],
                              [a[2], b[2], -n[2]]])
    ez: np.ndarray = np.array([0.0, 0.0, 1.0])
    s: np.ndarray = -sdd*np.matmul(np.linalg.inv(M), ez)

    return s[2]

def make_solid_angle_coverage_correction_factors(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, shortest_dist_to_detector: float
    ) -> np.ndarray:
    """Makes an array with the correction factor for solid angle coverage of each pixel. 

    Based on Equation 28 in Pauw J Phys.: Condens. Matter 25, 383201. 
    DOI: 10.1088/0953-8984/25/38/383201. 

    Args:
        x: x coordinates of pixels in the lab system. 
        y: y coordinates of pixels in the lab system. 
        z: z coordinates of pixels in the lab system. 
        shortest_dist_to_detector: shortest distance from the sample to the detector plane. 
            see :py:func:`calc_shortest_dist_to_detector`. 
    
    Returns:
        A 2D array of correction factors for each pixel. 
        The correction factor is normalized at the beam center. 
        The correction can be done by multiplying this array to the intensity array. 
    """

    # Lp = (x^2 + y^2 + z^2)^(1/2)
    # Lp^3 = (x^2 + y^2 + z^2)^(3/2)
    Lp3: np.ndarray = np.power(x*x + y*y + z*z, 3.0/2.0)

    return Lp3/np.power(shortest_dist_to_detector, 3.0)


def make_q(x: np.ndarray, y: np.ndarray, z: np.ndarray, wavelength: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes q vector. 
    
    By definition, 

    .. math::
       \\vec{q} = \dfrac{2 \\pi}{\\lambda}(\\vec{e}_\\mathrm{s} - \\vec{e}_\\mathrm{i})

    where :math:`\\lambda` is the wavelength, 
    :math:`\\vec{e}_\\mathrm{s}` is the basis vector along the scattered ray, 
    and :math:`\\vec{e}_\\mathrm{i}` is the basis vector along the incident ray. 

    Here, :math:`\\vec{e}_\\mathrm{i}` is fixed to (0, 0, 1). 
    Since the sample is placed at the origin, 
    
    .. math::
       \\vec{e}_\\mathrm{s} = \\dfrac{\\vec{r}}{|\\vec{r}|}

    where :math:`\\vec{r}` is the coordinate of the pixel in the lab system. 

    Args:
        x: x coordinates of pixels in the lab system. 
        y: y coordinates of pixels in the lab system. 
        z: z coordinates of pixels in the lab system. 
        wavelength: wavelength of the incident beam. 

    Returns:
        Three 2D arrays representing x, y, and z components of the q vector. 
    """

    ei_z: float = 1.0
    pre: float = 2.0*np.pi/wavelength

    Lp: float = np.sqrt(x*x + y*y + z*z)

    es_x: np.ndarray = x/Lp
    es_y: np.ndarray = y/Lp
    es_z: np.ndarray = z/Lp

    qx: np.ndarray = pre * es_x
    qy: np.ndarray = pre * es_y
    qz: np.ndarray = pre * (es_z - ei_z)

    return qx, qy, qz

def calc_polar_map(
    qx: np.ndarray, qy: np.ndarray, qz: np.ndarray, 
    mask: np.ndarray, 
    qmin: float, qmax: float,
    N_q: int, N_azi: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Calculates mapping to the polar coordinate system. 

    Args:
        qx: x component of q vector. 
        qy: y component of q vector. 
        qz: z component of q vector. 
        mask: mask array. 
        qmin: lower boundary of q. 
        qmax: upper boundary of q. 
        N_q: number of bins along the q axis. 
        N_azi: number of bins along the azimuthal axis. 
            360 deg is divided into `N_azi` sections. 
    
    Returns:
        Five numpy arrays, `map_q`, `map_azi`, `density`, `ax_q`, and `ax_azi`. 
    """

    assert qx.dtype == np.float64
    assert qy.dtype == np.float64
    assert qz.dtype == np.float64
    assert mask.dtype == np.uint8

    return _cpolarize.calc_polar_map(qx, qy, qz, mask, qmin, qmax, N_q, N_azi)


def circular_average(
    i: np.ndarray, e: np.ndarray, 
    map_q: np.ndarray, map_azi: np.ndarray, 
    density: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:

    assert i.dtype == np.float64
    assert e.dtype == np.float64
    assert map_q.dtype == np.int64
    assert map_azi.dtype == np.int64
    assert density.dtype == np.int64

    return _cpolarize.circular_average(i, e, map_q, map_azi, density)


if __name__ == "__main__":
    pass


# coding="utf-8"

from typing import List, IO, Tuple
import numpy as np

import math

def calc_polar_map(qx: np.ndarray, qy: np.ndarray, qz: np.ndarray, 
                   mask: np.ndarray, 
                   qmin: float, qmax: float, 
                   N_q: int, N_azi: int) \
                   -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    '''
    @param qx: x component of the q vector
    @param qy: y component of the q vector
    @param qz: z component of the q vector
    @param mask:
    @param qmax:
    @param N_q:
    @param N_azi:

    @return map_q, map_azi, density, ax_q, ax_azi:
    map_q: array of indices in q-axis and phi-axis.
    map_azi: array of indices in phi-axis.
    density: number of pixel in the section.
    ax_q: q-axis.
    ax_azi = phi-axis.
    '''

    q: float = 0.0
    azi: float =0.0

    map_q: np.ndarray = np.zeros_like(qx, dtype=int)
    map_azi: np.ndarray = np.zeros_like(qx, dtype=int)
    density: np.ndarray = np.zeros((N_azi, N_q), dtype=int)

    height: int = qx.shape[0]
    width: int = qx.shape[1]

    i_q: int = 0
    i_azi: int = 0

    delta_q: float = (qmax-qmin)/float(N_q)
    delta_azi: float = 2.0*np.pi/float(N_azi)

    ax_q: np.ndarray = qmin + (np.arange(N_q).astype(float) + 0.5) * delta_q
    ax_azi: np.ndarray = (np.arange(N_azi).astype(float) + 0.5) * delta_azi

    j: int = 0
    k: int = 0

    for j in range(height):
        for k in range(width):
            if (mask[j,k] == 1):
                map_q[j,k] = -1
                map_azi[j,k] = -1
                continue

            q = math.sqrt(qx[j,k]*qx[j,k] + qy[j,k]*qy[j,k] + qz[j,k]*qz[j,k]) # q = |q|
            i_q = int(math.floor((q-qmin)/delta_q)) # get index (fast)
            if (i_q >= N_q) or (q < qmin):      # if the index is out of bound,
                map_q[j,k] = -1
                map_azi[j,k] = -1
                continue

            if qx[j,k] == 0.0:           # very special cases
                if qy[j,k] > 0.0:
                    azi = np.pi/2.0
                if qy[j,k] < 0.0:
                    azi = 3.0*np.pi/2.0
                else:
                    azi = 0.0
            else:
                azi = math.atan(qy[j,k]/qx[j,k])       # azimuthal angle of the q vector projected to the x-y plane
                if qx[j,k] < 0.0:                  # convert limit of azi to [0, 2pi)
                    azi = azi + np.pi
                elif qy[j,k] < 0.0:
                    azi = azi + 2.0*np.pi
            i_azi = int(math.floor(azi/delta_azi))
            if i_azi >= N_azi:               # if the index is out of bound,
                map_q[j,k] = -1
                map_azi[j,k] = -1
                continue

            map_q[j,k] = i_q
            map_azi[j,k] = i_azi
            density[i_azi, i_q] = density[i_azi, i_q] + 1 # count up only for non-masked & valid pixels
    
    return map_q, map_azi, density, ax_q, ax_azi


def circular_average(i: np.ndarray, e: np.ndarray, 
                     map_q: np.ndarray, map_azi: np.ndarray, 
                     density: np.ndarray) \
                     -> Tuple[np.ndarray, np.ndarray]:
    '''
    Do circular-average using given maps.

    @param i: scattered intensity.
    @param map_q: indices in the q-axis.
    @param map_azi: indices in the phi-axis.
    @param density: number of pixels in the section.

    @note: Masking should be done beforehand, i.e., either in calc_q_vector() or calc_polar_map().

    @return ave_i, ave_e: 2d arrays of the reduced intensity and its error.
    '''

    ave_i: np.ndarray = np.zeros_like(density, dtype=float)
    ave_e2: np.ndarray = np.zeros_like(density, dtype=float)
    densityf: np.ndarray = density.astype(float) # float version of density
    densityf2: np.ndarray = np.power(densityf, 2)

    height: int = i.shape[0]
    width: int = i.shape[1]

    j: int = 0
    k: int = 0

    i_q: int = 0
    i_azi: int = 0

    # N_q: int  = densityf.shape[1]
    # N_azi: int = densityf.shape[0]

    # intensities
    for j in range(height):
        for k in range(width):
            i_q = map_q[j,k]
            i_azi = map_azi[j,k]

            if (i_q == -1) or (i_azi == -1) or (density[i_azi,i_q] == 0.0): # these pixels are masked
                pass
            else:
                try:
                    ave_i[i_azi, i_q] += i[j,k]/densityf[i_azi,i_q]          # normalized by the number of pixels
                    ave_e2[i_azi, i_q] += e[j,k]*e[j,k]/densityf2[i_azi,i_q] # from the law of error propagation
                except ZeroDivisionError:
                    pass

    return ave_i, np.sqrt(ave_e2)

if __name__ == "__main__":
    pass


# coding="utf-8"

from sgt import pilatus

if __name__ == "__main__":

    arr = pilatus.ReadTiff("test data PF/AgBh.tif")

    from matplotlib import pyplot

    pyplot.imshow(arr)

    pyplot.show()


# coding=utf-8
'''
Created on 2017/11/17

@author: snakagawa
'''

import struct
import numpy as np
from typing import Tuple, IO

def _splitnull(c_string: str) -> str:
    """
    Converts null terminated string to a string. 
    """
    buf = c_string.split("\0")
    return buf[0]

DTYPES: dict = {1: "B", 
                2: "s", 
                3: "H", 
                4: "L", 
                6: "b", 
                7: "c", 
                8: "h", 
                9: "l", 
                11: "f", 
                12: "d"}

def ReadTiff(fp: str|IO) -> np.ndarray:
    """Reads a .tif file from PILATUS detector and returns the 2-d data as a numpy array. 

    Args:
        fp: file-like or path to the file. 
    
    Returns:
        A 2D array of the int type. 

    Example:
        >>> arr = pilatus.ReadTiff("AgBh.tif")
        >>> print(arr.shape, arr.dtype)
    """
    f: IO|None = None
    close: bool = True
    byteorder: str = ""
    version: int = 0
    ifd_offset: int = 0
    entry_count: int = 0
    tag_dict: dict = {}
    entry_number: int = 0
    
    tag_code: int = 0
    dtype_code: int = 0
    count: int = 0
    data_str: str = ""
    data_offset: int = 0
    c_offset: int = 0

    dtype: str = ""
    numer: int = 0
    denom: int = 0
    data: float = 0.0

    dformat: str = ""
    dlength: int = 0


    ImageWidth: int = 0
    ImageHeight: int = 0
    BitsPerSample: Tuple[int, ...] = ()
    Compression: int = 0
    PhotometricInterpretation: int = 0
    StripOffsets: Tuple[int, ...]
    RowsPerStrip: int = 0
    StripBytesCounts: Tuple[int, ...] = ()
    XResolution: float = 0.0
    YResolution: float = 0.0
    NewSubfileType: int = 0
    DateTime: bytes = b""

    BitsPerSampleSingle: int = 0
    ImageDataOffset: int = 0
    ImageBytesCount: int = 0

    #------#
    # Open #
    #------#
    if isinstance(fp, str):
        f = open(fp, "rb")
        close = True
    else:
        f = fp
        close = False
    
    # f.seek(0)
    
    #--------#
    # Header #
    #--------#
    try:
        byteorder = {b'II': '<', b'MM': '>'}[f.read(2)]
    except KeyError:
        raise IOError("not a valid TIFF file")
    
    version, = struct.unpack(byteorder+"H", f.read(2))
    if version != 42:
        raise IOError("not a valid TIFF file")
    
    ifd_offset, = struct.unpack(byteorder+"l", f.read(4))
    
    #-----#
    # IFD #
    #-----#
    
    ### SEEK
    f.seek(ifd_offset)
    # number of entries
    entry_count, = struct.unpack(byteorder+"H", f.read(2))
    tag_dict = {}
    for entry_number in range(entry_count):
        
        tag_code, dtype_code, count, data_str = struct.unpack(byteorder+"HHl4s", f.read(12))
        # special data types (RATIONAL and SRATIONAL)
        if dtype_code in (5, 10) and count == 1:
            data_offset, = struct.unpack(byteorder+"L", data_str)
            ### POP
            c_offset = f.tell()
            f.seek(data_offset)
            if dtype_code == 5:
                dtype = "R"
                numer, denom = struct.unpack(byteorder+"LL", f.read(8))
            else:
                dtype = "S"
                numer, denom = struct.unpack(byteorder+"lL", f.read(8))
            data = float(numer)/denom    
            f.seek(c_offset)
            ### PUSH
        else:
            try:
                dtype = DTYPES[dtype_code]
            except KeyError:
                raise IOError("Contains unknown or unsupported data type")
            
            dformat = "{0}{1}{2}".format(byteorder, count, dtype)
            dlength = struct.calcsize(dformat)
            if dlength <= 4:
                data = struct.unpack(dformat, data_str[:dlength])
                data_offset = 0
            else:
                data_offset, = struct.unpack(byteorder+"L", data_str)
                ### POP
                c_offset = f.tell()
                f.seek(data_offset)
                data = struct.unpack(dformat, f.read(dlength))
                f.seek(c_offset)
                ### PUSH
                if type(data[0]) == str:
                    data = _splitnull(data[0])
        
        #print tag_code, dtype, count, data
        
        tag_dict[tag_code] = data
        
    #------------#
    # Check tags #
    #------------#
    try:
        ImageWidth    = tag_dict[256][0]
        ImageHeight   = tag_dict[257][0]
        BitsPerSample = tag_dict[258]
        Compression   = tag_dict[259][0]
        PhotometricInterpretation = tag_dict[262][0]
        StripOffsets  = tag_dict[273]
        RowsPerStrip  = tag_dict[278][0]
        StripBytesCounts = tag_dict[279]
        XResolution   = tag_dict[282]
        YResolution   = tag_dict[283]
        
        NewSubfileType = tag_dict[254][0]
        #ImageDescription = tag_dict[270][0]
        #Model         = tag_dict[272][0]
        #Software      = tag_dict[305]
        DateTime      = tag_dict[306][0]
        #Artist        = tag_dict[315]
        
        # print("ImageWidth", ImageWidth, type(ImageWidth))
        # print("ImageHeight", ImageHeight, type(ImageHeight))
        # print("BitsPerSample", BitsPerSample, type(BitsPerSample))
        # print("Compression", Compression, type(Compression))
        # print("PhotometricInterpretation", PhotometricInterpretation, type(PhotometricInterpretation))
        # print("StripOffsets", StripOffsets, type(StripOffsets))
        # print("RowsPerStrip", RowsPerStrip, type(RowsPerStrip))
        # print("StripBytesCounts", StripBytesCounts, type(StripBytesCounts))
        # print("XResolution", XResolution, type(XResolution))
        # print("YResolution", YResolution, type(YResolution))
        # print("NewSubfileType", NewSubfileType, type(NewSubfileType))
        # print("DateTime", DateTime, type(DateTime))

    except KeyError:
        raise IOError("Missing tag(s)")
    
    if len(BitsPerSample) > 1:
        raise IOError("Unsupported tiff (color): BitsPerSample=" + ", ".join([str(e) for e in BitsPerSample]))
    else:
        BitsPerSampleSingle = BitsPerSample[0]
    
    if Compression != 1:
        raise IOError("Unsupported tiff (compressed)")
    
    if PhotometricInterpretation != 1:
        raise IOError("Unsupported tiff (wrong photometric interpretation)")
    
    if len(StripOffsets) != 1:
        raise IOError("Unsupported tiff (multiple strips)")
    else:
        ImageDataOffset = StripOffsets[0]
    
    ImageBytesCount = StripBytesCounts[0]
    
    if NewSubfileType != 0:
        raise IOError("Unsupported tiff (multipage or mask)")
        
    #-----------#
    # Load data #
    #-----------#
    f.seek(ImageDataOffset)
    PixelNumber = ImageWidth * ImageHeight
    if BitsPerSampleSingle == 32:
        flat = np.fromfile(f, dtype=np.int32, count=PixelNumber)
        image = np.reshape(flat, (ImageHeight, ImageWidth), order="C")
    
    if close:
        f.close()
    
    return image

# def LoadPilatusTiff(fp, key_dest="intensity"):
#     im = ReadGrayTiff(fp)
#     wim = snpl.image.NpzImage()
#     wim.append_layer(key_dest, im)
#     return wim
# 
# def get_mask_NpzImage(fp, empty=False):
#     im = ReadGrayTiff(fp)
#     wim = snpl.image.NpzImage()
#     mask = np.zeros_like(im, dtype=np.uint8)
#     if not empty:
#         mask[np.where(im<0)] = 1
#     wim.append_layer("mask", mask)
#     return wim

if __name__ == '__main__':
    im = ReadTiff("../test/scatter/AgBh.tif")


# coding="utf-8"

import sgt
from matplotlib import pyplot
import snpl

import numpy as np

if __name__ == "__main__":
    g = sgt.geometry()
    g.load_specs("test data PF/Geometry_AgBh.txt")
    g.refresh_q()

    mask = snpl.image.NpzImage("test data PF/Mask_EmptyBeam_ManuallyEdited.npz").get_layer("mask")
    i = sgt.pilatus.ReadTiff("test data PF/AgBh.tif").astype(np.float64)
    e_i = np.sqrt(i)

    g.mask = mask
    g.refresh_polar_map()

    g.save_specs("test data PF/Geometry_AgBh_saved.txt")

    g2 = sgt.geometry()
    g2.load_specs("test data PF/Geometry_AgBh_saved.txt")
    g2.mask = mask
    g2.refresh_q()
    g2.refresh_polar_map()

    av, e_av = g.circular_average(i, e_i)

    maskp = np.full_like(mask, np.nan, dtype=float)
    maskp[mask==1] = 1.0

    pyplot.pcolormesh(g.x, g.y, i, shading="nearest")
    pyplot.pcolormesh(g.x, g.y, maskp, shading="nearest", cmap="Greys", vmin=0, vmax=1)

    pyplot.gca().set_aspect("equal")

    pyplot.savefig("test_geometry.png", dpi=300)
    pyplot.cla()
    pyplot.clf()

    q, azi = np.meshgrid(g.ax_q, g.ax_azi)

    print(q)

    pyplot.pcolormesh(q, azi, av, shading="nearest", vmax=10000.0)
    pyplot.savefig("test_geometry_averaged.png", dpi=300)


# coding="utf-8"

import numpy as np
from sgt import crystal
from matplotlib import pyplot


if __name__ == "__main__":
    
    a = 1.5
    b = 2.0
    c = 4.0

    al = 100.0
    be = 110.0
    ga = 120.0

    avec, bvec, cvec = crystal.make_abc_vectors(a, b, c, al, be, ga)

    ax = pyplot.figure().add_subplot(projection="3d")
    ax.set_box_aspect((1,1,1))
    ax.set_xlim(-5.0, 5.0)
    ax.set_ylim(-5.0, 5.0)
    ax.set_zlim(-5.0, 5.0)
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    def plot_vector(ax, vec, **kwargs):
        return ax.plot([0.0, vec[0]], [0.0, vec[1]], [0.0, vec[2]], **kwargs)
    
    print("a", np.linalg.norm(avec))
    print("b", np.linalg.norm(bvec))
    print("c", np.linalg.norm(cvec))
    print("a-b", np.rad2deg(np.arccos(np.dot(avec, bvec)/a/b)))
    print("b-c", np.rad2deg(np.arccos(np.dot(bvec, cvec)/b/c)))
    print("c-a", np.rad2deg(np.arccos(np.dot(cvec, avec)/c/a)))

    plot_vector(ax, [1.0, 0.0, 0.0], ls="-", color="r", alpha=0.5)
    plot_vector(ax, [0.0, 1.0, 0.0], ls="-", color="g", alpha=0.5)
    plot_vector(ax, [0.0, 0.0, 1.0], ls="-", color="b", alpha=0.5)

    plot_vector(ax, avec, ls="-", color="r", lw=2)
    plot_vector(ax, bvec, ls="-", color="g", lw=2)
    plot_vector(ax, cvec, ls="-", color="b", lw=2)

    pyplot.show()


# coding="utf-8"
from typing import List, Tuple, Callable, Any
import sgt
from sgt import _crefine
import numpy as np
import scipy.optimize

def score_by_rings(
    geom: sgt.geometry, 
    i: np.ndarray, 
    qtargets: List[float], 
    qlimwidths: List[float], 
    qfuncwidths: List[float], 
    verbose: bool = False
    ) -> float:
    """Evaluates the score of the geometry. 
    """

    ret: float = 0.0
    qtarget: float = 0.0
    qlimwidth: float = 0.0
    qfuncwidth: float = 0.0
    qmin: float = 0.0
    qmax: float = 0.0
    
    # for every ring, 
    for qtarget, qlimwidth, qfuncwidth in zip(qtargets, qlimwidths, qfuncwidths):
        
        # q limits of the ring
        qmin = qtarget - qlimwidth
        qmax = qtarget + qlimwidth

        # calculate the score and add it to ret
        ret = ret + _crefine.proximity(
            i.astype(np.float64), 
            geom.qx, geom.qy, geom.qz, 
            geom.solid_angle_factor, 
            geom.mask, qtarget, qmin, qmax, qfuncwidth)

    # return 1.0/ret
    if verbose:
        print(-ret)

    return -ret

def refine(
    geom: sgt.geometry, 
    score_func: Callable[..., float], 
    score_args: Tuple[Any, ...], 
    refine_specnames: Tuple[str, ...], 
    score_requires_polarmap: bool = False, 
    optimize_method: str = "Nelder-Mead", 
    callback: Callable[[int, float], None]|None = None
    ) -> None:
    """Refines a geometry. 

    The parameters in ``specs`` of ``geom``
    are optimized so that ``score_func`` is minimized. 

    The ``score_func`` should be a callable that takes 
    a ``geometry`` object as the first argument. 
    It will be called as
    ```python
    score_func(geom, *score_args)
    ```
    during the optimization. 

    ``geom`` will be updated on every iteration of optimization. 
    The names of the parameters in ``specs`` to be optimized 
    must be specified by ``refine_specnames``. 

    Args:
        geom: A geometry object to be optimized. 
        score_func: A scoring function with the call signature 
            ``score_func(geom, *score_args)``. 
        score_args: Args for ``score_func``. 
        refine_specnames: Names of the parameters to be optimized. 
        score_requires_polarmap: If True, ``geom.refresh_polarmap()`` 
            will be called on every iteration. 
            If False, only ``geom.refresh_q()`` will be called. 
        optimize_method: Optimization method to be passed to 
            ``scipy.optimize.minimize``. 
        callback: A callback function to be called 
    
    """

    ncalls: int = 0
    score: float = 0.0

    # initial values of the parameters to be optimized
    p0: List[Any] = [geom.specs[k] for k in refine_specnames]

    # define scoring function that takes the parameters to be optimized as the args
    def _get_score(params: List[Any]) -> float:
        nonlocal ncalls

        # update the geometry
        geom.specs.update({k: v for k, v in zip(refine_specnames, params)})
        geom.refresh_q()
        if score_requires_polarmap:
            geom.refresh_polar_map()

        score = score_func(geom, *score_args)
        ncalls = ncalls + 1
        callback(ncalls, score)

        return score

    res = scipy.optimize.minimize(_get_score, p0, method=optimize_method)

if __name__ == "__main__":
    pass


# coding="utf-8"

from setuptools import setup, Extension
from numpy import get_include

setup_args = dict(
    ext_modules = [
        Extension(
            'sgt._cpolarize',
            ['src/sgt/_cpolarize.c'],
            include_dirs = [get_include()]
        ), 
        Extension(
            'sgt._crefine',
            ['src/sgt/_crefine.c'],
            include_dirs = [get_include()]
        )
    ]
)
setup(**setup_args)


# coding="utf-8"

from typing import Tuple
import numpy as np

def make_primitive_vectors(
    a: float, 
    b: float, 
    c: float, 
    alpha_deg: float=90.0, 
    beta_deg: float=90.0, 
    gamma_deg: float=90.0
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    alpha: float = np.deg2rad(alpha_deg)
    beta: float = np.deg2rad(beta_deg)
    gamma: float = np.deg2rad(gamma_deg)

    cos_al: float = np.cos(alpha)
    sin_al: float = np.sin(alpha)

    cos_be: float = np.cos(beta)
    sin_be: float = np.sin(beta)

    cos_ga: float = np.cos(gamma)
    sin_ga: float = np.sin(gamma)

    avec: np.ndarray = np.array([a, 0.0, 0.0])
    bvec: np.ndarray = np.array([b*cos_ga, b*sin_ga, 0.0])

    c1: float = c*cos_be
    c2: float = c*(cos_al - cos_be*cos_ga)/sin_ga
    c3: float = np.sqrt(c*c - c1*c1 - c2*c2)

    cvec: np.ndarray = np.array([c1, c2, c3])

    return avec, bvec, cvec

def make_reciprocal_primitive_vectors(
    avec: np.ndarray, bvec: np.ndarray, cvec: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    pref: float = 2.0*np.pi*np.dot(avec, np.outer(bvec, cvec))

    return pref*np.outer(bvec, cvec), pref*np.outer(cvec, avec), pref*np.outer(avec, bvec)

def make_reciprocal_lattice_vector(
    h: int, k: int, l: int, 
    astar: np.ndarray, bstar: np.ndarray, cstar: np.ndarray
    ) -> np.ndarray:

    return h*astar + k*bstar + l*cstar

if __name__ == "__main__":
    pass